/*********************************************************************/
/* check_routes.c                                                    */
/* neblo_check_neighbours(), neblo_check_fingers():                  */
/* autonomous threads for routing tables maintenance.                */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>



/*
 *  Send the current list of neighbours to another peer
 *  (usually the predecessor).
 *  Returns 0 if everything OK, <0 if peer "dest" had troubles.
 */
static int
send_neighlist
( peer_t *dest, pthread_t th_id )
{

  ringofptrs_t         *alt_ring;
  peerptr_t            *tail,*ff;
  char                 *cmd;
  char                 *dest_keymod;
  char                 *dest_keyexp;
  peer_t               *f;
  char                 *neigh_keymod;
  char                 *neigh_keyexp;
  int                   rc;
  int                   i;


  /*
   *  Make a list of ptrs to current neighbours, so we don't need to assert
   *  a long-lasting lock over the neighbour list.
   */

  _do_(_calloc_(alt_ring,sizeof(ringofptrs_t)),!alt_ring,
       "send_neighlist(): alt_ring malloc");

  pthread_mutex_lock(&ring->lock);
  memcpy(&alt_ring->ring,ring,sizeof(ring_t));
  pthread_mutex_unlock(&ring->lock);

  alt_ring->peerlist = tail = NULL;

  forall_neighbours(f,
    _do_(_calloc_(ff,sizeof(peerptr_t)),!ff,"send_neighlist(): ff calloc");
    ff->f = f;
    if (f == ring->neigh_list)
      alt_ring->peerlist = tail = ff;
    else {
      tail->next = ff;
      tail = ff;
    }
    ff->next = NULL;
  );

  _do_(_malloc_(cmd,KEYBUFLEN+SLACK),!cmd,"send_neighlist(): cmd malloc");

  _do_(_BN_bn2hex_(dest_keymod,dest->pubkey->n),!dest_keymod,
       "send_neighlist(): dest_keymod alloc");
  _do_(_BN_bn2hex_(dest_keyexp,dest->pubkey->e),!dest_keyexp,
       "send_neighlist(): dest_keyexp alloc");

  /*  Dialog with peer "dest"; protect it with lock  */

  pthread_mutex_lock(&dest->lock);

  for (;;) {

    rc = tell_peer_nolock(dest,2,"neighlist",strlen("neighlist")+1,0,th_id);
    if (rc < 0) break;

#ifdef DEBUG_M
    S(fprintf(stderr,"[%lu]\t<< %s\n",th_id,"neighlist"));
#endif

    rc = listen_peer_nolock(dest,cmd,BUFLEN,TIMEOUT,0);
    if (rc < 0) break;

    if (!strncmp(cmd,"wait",strlen("wait"))) {
      /*  recipient not yet entered the ring.  quit, try later.  */
      rc = 0;
      break;
    }

    if (strncmp(cmd,"ok",strlen("ok"))) {
      rc = -1;
      break;
    }

#ifdef DEBUG_M
    S(fprintf(stderr,"[%lu]\t>> %s\n",th_id,cmd));
#endif

    /*  recipient ready to get the list  */

    for (ff=alt_ring->peerlist,i=0; ff!=NULL && i<MAXNEIGH; ff=ff->next, i++) {

      f = ff->f;

      if (!(f->status & (OK | TOCALL | PROBABLYDEAD))) {

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\tskipping %u.%u.%u.%u:%u, status %u...\n",
                  th_id,
                  ((unsigned char *)&f->addr.s_addr)[0],
                  ((unsigned char *)&f->addr.s_addr)[1],
                  ((unsigned char *)&f->addr.s_addr)[2],
                  ((unsigned char *)&f->addr.s_addr)[3],
                  f->port,f->status));
#endif

        continue;
      }

      sprintf(cmd,"%lu %u",(unsigned long)f->addr.s_addr,f->port);
      rc = tell_peer_nolock(dest,2,cmd,strlen(cmd)+1,0,th_id);
      if (rc < 0) break;

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\t<< \"neighbour %u.%u.%u.%u:%u\"\n",
                th_id,
                ((unsigned char *)&f->addr.s_addr)[0],
                ((unsigned char *)&f->addr.s_addr)[1],
                ((unsigned char *)&f->addr.s_addr)[2],
                ((unsigned char *)&f->addr.s_addr)[3],
                f->port));
#endif

      rc = listen_peer_nolock(dest,cmd,BUFLEN,TIMEOUT,0);
      if (rc < 0) break;

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\t>> %s\n",th_id,cmd));
#endif

      if (!strncmp(cmd,"ok",strlen("ok")))
        /*  recipient not interested in this neighbour, go to next one  */
        continue;

      if (strncmp(cmd,"pubkey",strlen("pubkey"))) {
        rc = -1;
        break;
      }

      /*
       *  recipient interested in this neighbour; announce that neighbour that
       *  recipient might ask it for friendship, then send pubkey.
       */

      rc = announce_friend(f,dest->addr,dest->port,dest_keymod,dest_keyexp,
                           th_id);
      /* don't care about neighbour troubles; "dest" will check on its own */

      _do_(_BN_bn2hex_(neigh_keymod,f->pubkey->n),!neigh_keymod,
           "send_neighlist(): neigh_keymod alloc");
      _do_(_BN_bn2hex_(neigh_keyexp,f->pubkey->e),!neigh_keyexp,
           "send_neighlist(): neigh_keyexp alloc");
      sprintf(cmd,"%s %s",neigh_keymod,neigh_keyexp);
      _OPENSSL_free_(neigh_keymod); _OPENSSL_free_(neigh_keyexp);

      rc = tell_peer_nolock(dest,2,cmd,strlen(cmd)+1,0,th_id);
      if (rc < 0) break;

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\t<< \"neighbour pubkey sent\"\n",th_id));
#endif

      rc = listen_peer_nolock(dest,cmd,BUFLEN,TIMEOUT,0);
      if (rc < 0) break;

      if (strncmp(cmd,"ok",strlen("ok"))) {
        rc = -1;
        break;
      }

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\t>> %s\n",th_id,cmd));
#endif

      /*  recipient ready to get next neighbour of list, continue loop  */
    
    }

    /*  Abnormally quit loop?  */
    if (rc < 0) break;

    /*  End of neigh list  */
    rc = tell_peer_nolock(dest,2,"end",strlen("end")+1,0,th_id);
    if (rc < 0) break;

#ifdef DEBUG_M
    S(fprintf(stderr,"[%lu]\t<< end\n",th_id));
#endif

    rc = 0;
    break;

  }

  /*  End of dialog.  */
  pthread_mutex_unlock(&dest->lock);

  if (rc < 0) goto abort;

exit:

  for (ff=alt_ring->peerlist; ff!=NULL;) {
    tail = ff->next;
    _free_(ff);
    ff = tail;
  }
  _free_(alt_ring);
  _free_(cmd);
  _OPENSSL_free_(dest_keymod);
  _OPENSSL_free_(dest_keyexp);
  return rc;

abort:

  rc = -1;
  goto exit;

}



/*
 * neblo_check_neighbours.
 *
 * Periodical maintenance and liveness check for neighbours, friends and
 * predecessor.  Check successor more often than others.
 *
 * Tries to garbage-collect DEAD entries from such tables.
 * Tries to seal the ring after failure of successor.
 *
 * After maintenance, the current neighbour list is backward propagated
 * along the ring.
 * 
 * Does not return.
 *
 */
void *
neblo_check_neighbours
( void *data ) 
{
  
  pthread_t             th_id;
  char                 *cmd,*rsp;
  int                   rc;
  peer_t               *f,*ff,*p;
  int                   timeout;
  ULL                   intvl;
  neblo_addr            interval;
  neblo_addr            old_up_bound;
  struct timeval        t1,t2,t3,tv,tv2;
  int			check_all;
  
  
  th_id = *(pthread_t *)data;

  _do_(_malloc_(cmd,KEYBUFLEN+SLACK),!cmd,"neblo_check_friends(): cmd malloc");
  _do_(_malloc_(rsp,KEYBUFLEN+SLACK),!rsp,"neblo_check_friends(): rsp malloc");

  gettimeofday(&t1,NULL);
  t3 = t1;

  for (;;) {
    
    /*
     *  Run periodically, insensitive to the time spent to carry out the
     *  checks.
     */
    gettimeofday(&t2,NULL);
    if ((unsigned int)(t2.tv_sec - t1.tv_sec) < CHK_PERIOD_SUCCESSOR)
      sleep(CHK_PERIOD_SUCCESSOR - (t2.tv_sec - t1.tv_sec));
    gettimeofday(&t1,NULL);
    if ((unsigned int)(t1.tv_sec - t3.tv_sec) < CHK_PERIOD_NEIGHS)
      check_all = 0;
    else {
      t3 = t1;
      check_all = 1;
    }

#ifdef DEBUG_R
    S(
    fprintf(stderr,"[%lu]\tneblo_check_neighbours(%d)\n",th_id,check_all);
    fprintf(stderr,"[%lu]\tlocal keys: %016llx -- %016llx\n",
            th_id,ring->my_id,ring->up_bound);
    fprintf(stderr,"[%lu]\tcurrent neighbour list:\n",th_id);
    );
    forall_neighbours(f,
      if (!f->self)
        S(
        fprintf(stderr,"[%lu]\t%u.%u.%u.%u:%u (%u)\n",
                th_id,
                ((unsigned char *)&f->addr.s_addr)[0],
                ((unsigned char *)&f->addr.s_addr)[1],
                ((unsigned char *)&f->addr.s_addr)[2],
                ((unsigned char *)&f->addr.s_addr)[3],
                f->port,f->status);
        );
    );
    S(fprintf(stderr,"[%lu]\tcurrent friend list:\n",th_id));
    forall_friends(f,
      S(
      fprintf(stderr,"[%lu]\t%u.%u.%u.%u:%u (%u) %s\n",
              th_id,
              ((unsigned char *)&f->addr.s_addr)[0],
              ((unsigned char *)&f->addr.s_addr)[1],
              ((unsigned char *)&f->addr.s_addr)[2],
              ((unsigned char *)&f->addr.s_addr)[3],
              f->port,f->status,
              f->from_predecessor ? "PRED" : f->from_successor ? "SUCC" : " ");
      );
    );
#endif


    /* maintenance, step 1: liveness check for neighbours, friends, predec. */

    sprintf(cmd,"ping");

    p = NULL;

    if (check_all) {
      /* initiate predecessor liveness check, if not already DEAD */
      p = predecessor();
      if (p != NULL)
        rc = tell_peer(p,2,cmd,strlen(cmd)+1,0,th_id);
    }

    /* take current time for calibrating subsequent timeouts */
    gettimeofday(&tv,NULL);

    /*
     * Initiate neighbour liveness checks.
     * Trouble: we should raise lock on neighlist, but this check takes a
     * long time to complete and we don't want block routing for this long.
     * So, we don't raise ring->lock here.
     * peppe: potentially dangerous!!!
     */

    forall_neighbours_nolock(f,
      if (f->self) continue;
      if (f->status & DEAD) continue;
      if (!check_all && !is_successor_nolock(f)) continue;
      rc = tell_peer(f,is_successor_nolock(f) ? 1 : 0,cmd,strlen(cmd)+1,0,
                     th_id);
    );

    /*
     * we must give precedence to checking predecessor w.r.t. neighbours,
     * to ensure the predecessor/successor connection pairs will
     * match also in "small" rings (rings in which the neigh list covers all
     * partecipants).
     * so, check predecessor at first (if any selected).
     */

    for (;;) {

      /*  if no predecessor, skip the check.  */
      if (p == NULL) break;

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\tcheck predecessor (%u)...\n",th_id,p->port));
#endif

      bzero(rsp,4);
      gettimeofday(&tv2,NULL);
      timeout = min(1,TIMEOUT - ((int)tv2.tv_sec - (int)tv.tv_sec));
      rc = listen_peer(p,rsp,strlen("ping")+1,timeout,0);
      if (rc >= 0) {
        rc = strncmp(rsp,cmd,strlen(cmd));
        rc = -(rc != 0);  /*  -1 iff strings differ from each other  */
      }
      if (rc >= 0) {
        /* predecessor OK */

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\tOK\n",th_id));
#endif

        break;
      }
      if (rc == -2) {
        /* predecessor is dead and its status has been set to DEAD */

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        break;
      }
      /* predecessor has troubles but unsure is dead */
      if (!(p->status & PROBABLYDEAD)) {
        /*  countdown to death begins...  */
        p->status = PROBABLYDEAD;
        p->deadline = PEER_DEADLINE;
        if (p->deadline) {

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\tPROBABLYDEAD, deadline %u\n",
                  th_id,p->deadline));
#endif

        } else {
          p->status = DEAD;

#ifdef DEBUG_M
          S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        }
      } else {
        /*  countdown to death proceeds...  */
        if (!--p->deadline) {
          /* troubles for too long a time. assume it is dead */
          p->status = DEAD;

#ifdef DEBUG_M
          S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        }

#ifdef DEBUG_M
        else
          S(fprintf(stderr,"[%lu]\tPROBABLYDEAD, deadline %u\n",
                    th_id,p->deadline));
#endif

      }

      break;

    }

    /*
     * Finalize neighbour checks.
     * Cannot use forall_neighbours_nolock() macro because the body has to
     * contain some #ifdef directives, and some compilers dislike that.
     * Trouble: we should raise lock on neighlist, but this check takes a
     * long time to complete and we don't want block routing for this long.
     * So, we don't raise ring->lock here.
     * peppe: potentially dangerous!!!
     */

    for (f=ring->neigh_list; f!=NULL; f=f->next) {

      if (f->self) continue;
      if (f->status & DEAD) continue;
      if (!check_all && !is_successor_nolock(f)) continue;

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\tcheck neighbour %u.%u.%u.%u:%u...\n",
              th_id,
              ((unsigned char *)&f->addr.s_addr)[0],
              ((unsigned char *)&f->addr.s_addr)[1],
              ((unsigned char *)&f->addr.s_addr)[2],
              ((unsigned char *)&f->addr.s_addr)[3],
              f->port));
#endif

      bzero(rsp,4);
      gettimeofday(&tv2,NULL);
      timeout = min(1,TIMEOUT - ((int)tv2.tv_sec - (int)tv.tv_sec));
      rc = listen_peer(f,rsp,strlen("ping")+1,timeout,0);
      if (rc >= 0) {
        rc = strncmp(rsp,cmd,strlen(cmd));
        rc = -(rc != 0);  /*  -1 iff strings differ from each other  */
      }
      if (rc >= 0) {
        /* neighbour OK, go next */

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\tOK\n",th_id));
#endif

        continue;
      }
      if (rc == -2) {
        /* neighbour is dead and its status has been set to DEAD. go next */

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        continue;
      }
      /* neighbour has troubles but unsure is dead */
      if (f->status != PROBABLYDEAD) {
        /*  countdown to death begins...  */
        f->status = PROBABLYDEAD;
        f->deadline = PEER_DEADLINE;
        if (f->deadline) {

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\tPROBABLYDEAD, deadline %u\n",
                  th_id,f->deadline));
#endif

        } else {
          f->status = DEAD;

#ifdef DEBUG_M
          S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        }
      } else {
        /*  countdown to death proceeds...  */
        if (!--f->deadline) {
          /* troubles for too long a time. assume it is dead */
          f->status = DEAD;

#ifdef DEBUG_M
          S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        }

#ifdef DEBUG_M
        else
          S(fprintf(stderr,"[%lu]\tPROBABLYDEAD, deadline %u\n",
                    th_id,f->deadline));
#endif

      }

    };   /*  for all neighbours  */

    if (!check_all) goto after_clean;

    /*
     * check friends.
     * dead friends: timed out incoming connections.  do_serv() uses _close_()
     * to disposes timed out input sockets, so we just look at csock ptrs.
     */

    forall_friends(f, if (*f->in_cs == NULL) f->status = DEAD;);


    /* maintenance, step 2: remove neighbours and friends marked DEAD */

    pthread_mutex_lock(&clean_lock);

    if (!dont_clean) {

#ifdef DEBUG_M
      S(
      fprintf(stderr,"[%lu]\tneblo_check_neighbours(), step 2\n",th_id);
      fprintf(stderr,"[%lu]\tneighbour list:\n",th_id);
      );
      forall_neighbours(f,
        if (!f->self)
          S(
          fprintf(stderr,"[%lu]\t%u.%u.%u.%u:%u (%u)\n",
                  th_id,
                  ((unsigned char *)&f->addr.s_addr)[0],
                  ((unsigned char *)&f->addr.s_addr)[1],
                  ((unsigned char *)&f->addr.s_addr)[2],
                  ((unsigned char *)&f->addr.s_addr)[3],
                  f->port,f->status);
          );
      );
      S(fprintf(stderr,"[%lu]\tfriend list:\n",th_id));
      forall_friends(f,
        S(
        fprintf(stderr,"[%lu]\t%u.%u.%u.%u:%u (%u) %s\n",
                th_id,
                ((unsigned char *)&f->addr.s_addr)[0],
                ((unsigned char *)&f->addr.s_addr)[1],
                ((unsigned char *)&f->addr.s_addr)[2],
                ((unsigned char *)&f->addr.s_addr)[3],
                f->port,f->status,
                f->from_predecessor ? "PRED" : f->from_successor ? "SUCC" : " ");
        );
      );
#endif

      /*
       * Purge the neighbour/friend lists from DEAD elements,
       * if no protocol cmds are in progress which require
       * access to the routing tables (all except "announce" and "ping").
       * Here we must raise all necessary locks on the lists.
       */

      if (ring->neigh_list != NULL) {
        /* leave successor in list even if DEAD; might try to reuse soon... */
        purge_peer_list(&ring->neigh_list->next,&ring->neigh_list_tail,
                        &ring->lock);
        if (ring->neigh_list_tail == NULL)
          ring->neigh_list_tail = ring->neigh_list;
        /*
         * ...but dispose it anyway if no other useful neighbour left,
         * as in such case we have no chance to find any new successor
         * (we have to re-connect to ring).
         * "self" is not a useful neighbour ;-)
         */
        f = ring->neigh_list;
        if ((f->next == NULL || f->next->self)
            && (f->status & DEAD) && !pend_read_ref(f)) {
          ring->neigh_list = ring->neigh_list->next;
          if (ring->neigh_list == NULL)
            ring->neigh_list_tail = NULL;
          clean_peer_desc(f);
          _free_(f);
        }
      }

      purge_peer_list(&ring->friend_list,&ring->friend_list_tail,
                      &ring->friend_lock);

    }

    pthread_mutex_unlock(&clean_lock);

after_clean:

    if (successor() == NULL) {
      /* this peer is now isolated */
      if (!MAXSEEDS) {
        /* this peer is a seed and can run in isolation */
        /* all addresses are initially allocated to this peer */
        if (check_all) S(fprintf(stderr,"[%lu]\tpeer isolated\n",th_id));
        ring->up_bound = ring->my_id - 1;
      } else {
//peppe
#if 0
        S(fprintf(stderr,"[%lu]\tpeer isolated.  Re-join...\n",th_id));
        /* try to re-join this ring; will check at next round */
        neblo_submit();
#else
        S(fprintf(stderr,"[%lu]\tpeer isolated\n",th_id));
#endif
      }
      continue;
    }

    /* from this point on, routing table is ensured to be non-empty */

    /*
     * step 3: if successor is DEAD, find the new successor among neighbours,
     * then restore data integrity.
     * "self" is not a useful neighbour ;-)
     */
    ff = successor();
    if (!(ff->status & DEAD))
      goto step4;

    f = NULL;
    forall_neighbours(fff,
      if (!fff->self && fff != ff && fff->status & (OK|TOCALL|PROBABLYDEAD)) {
        f = fff; break;
      }
    );
    if (f == NULL) goto step4;

#ifdef DEBUG_M
    S(fprintf(stderr,
              "[%lu]\tneblo_check_neighbours(), step 3: find new successor\n",
              th_id));
#endif

#if 0
    /*  Forbid use of successor as long as its status is subject to change  */
   pthread_mutex_lock(&ff->lock);
   ff->status = DEAD;
#endif

    /*  Create handle for this req, it might be served in a recursive way  */
    cmd_handle_lock(&ff->handle);
    cmd_handle_set(&ff->handle);

    sprintf(cmd,"findsucc %u %lu %u %s %s",
            cmd_handle_val(&ff->handle),(unsigned long)my_ipaddr.s_addr,PORT,
            pubkeymod,pubkeyexp);

    /*  Begin a dialog with neighbour "f"  */
    pthread_mutex_lock(&f->lock);

    for (;;) {

      rc = tell_peer_nolock(f,0,cmd,strlen(cmd)+1,0,th_id); if (rc < 0) break;

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\t<< %s\n",th_id,cmd));
#endif

      rc = listen_peer_nolock(f,cmd,BUFLEN,TIMEOUT,0); if (rc < 0) break;

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\t>> %s\n",th_id,cmd));
#endif

      if (!strncmp(cmd,"forwarded",strlen("forwarded"))) {
        /*
         *  The "findsucc" will be answered by another peer ("newsucc" cmd).
         *  No race condition between this thread and the "newsucc" management,
         *  as we still hold the cmd_handle lock here.
         */

        /*  End of dialog.  */
        pthread_mutex_unlock(&f->lock);

        cmd_handle_unlock(&ff->handle);
        pthread_mutex_unlock(&ff->lock);

        rc = 0;
        break;
      }

      if (strncmp(cmd,"ok",strlen("ok"))) {
        /*  Wrong cmd from new successor, or successor actually not dead.   */
        rc = -1;
        break;
      }

      /* "f" is our new successor, expect an incoming conn from it.  */
      /* No race condition, because we still have to say "ok" to "f" */ 
      rc = insert_expd_reqst(f->addr,f->port,2 /* from successor */,
                             f->pubkey,th_id);
      if (rc < 0) break;

      rc = tell_peer_nolock(f,0,"ok",strlen("ok")+1,0,th_id);
      if (rc < 0) break;

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\t<< ok\n",th_id));
#endif

      /*
       *  The incoming conn from successor will be managed by a separate
       *  thread.  New successor tells us "end" once it has connected.
       */
      rc = listen_peer_nolock(f,cmd,BUFLEN,TIMEOUT,0); if (rc < 0) break;

      if (strncmp(cmd,"end",strlen("end"))) {
        rc = -1;
        break;
      }

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\t>> %s\n",th_id,cmd));
#endif

      /*
       *  New successor found, "f" is its descriptor.  Now we can purge the
       *  old, DEAD successor ("ff") from the neighbour list, and extend
       *  local addr interval.
       *  Actually, we prefer to simply move the old successor at the back of
       *  neighbour list; will be disposed at next round.
       */

      /*  End of dialog.  */
      pthread_mutex_unlock(&f->lock);

      cmd_handle_reset(&ff->handle);
      cmd_handle_unlock(&ff->handle);
#if 0
      pthread_mutex_unlock(&ff->lock);
#endif

      sscanf(cmd,"%*s %016llx",&intvl);
      interval = intvl;

      pthread_mutex_lock(&ring->lock);
      old_up_bound = ring->up_bound; 
      ring->up_bound = interval - 1;
      ring->neigh_list = f;
      ring->neigh_list_tail->next = ff;
      for (; ff->next != f; ff = ff->next);
      ff->next = NULL;
      ring->neigh_list_tail = ff;
      pthread_mutex_unlock(&ring->lock);

#ifdef DEBUG_D
      S(fprintf(stderr,"ciao restore %016llx-%016llx\n",
                ring->my_id,ring->up_bound));
#endif

      ask_refresh_addrs(old_up_bound+1,ring->up_bound,NEBLO_MAXCP,th_id);

      rc = 0;
      break;

    }

    if (rc < 0) {
      /*  Failed to accomplish "findsucc".  Retry next time.  */

      /*  End of dialog.  */
      pthread_mutex_unlock(&f->lock);

      cmd_handle_reset(&ff->handle);
      cmd_handle_unlock(&ff->handle);
#if 0
      ff->status = PROBABLYDEAD;
      ff->deadline = 1;
      pthread_mutex_unlock(&ff->lock);
#endif

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\tneblo_check_neighbours(), step 3: failure\n",
                th_id));
#endif

    }

step4:

    if (!check_all) goto done;

    /* step 4: backward-propagate new neighbour list */

#ifdef DEBUG_M
    S(fprintf(stderr,
              "[%lu]\tneblo_check_neighbours(), step 4: backward-propagating neigh list\n",
              th_id));
#endif

    f = predecessor();
    if (f != NULL) {
      rc = send_neighlist(f,th_id);

#ifdef DEBUG_M
      if (rc >= 0)
        S(fprintf(stderr,"[%lu]\tOK\n",th_id));
      else
        S(fprintf(stderr,"[%lu]\tfailed\n",th_id));
#endif

    }

done:

    /* Done */

#ifdef DEBUG_R
    S(
    fprintf(stderr,"[%lu]\tneblo_check_neighbours(), end:\n",th_id);
    fprintf(stderr,"[%lu]\tlocal keys: %016llx -- %016llx\n",
            th_id,(ULL)ring->my_id,(ULL)ring->up_bound);
    fprintf(stderr,"[%lu]\tnew neighbour list:\n",th_id);
    );
    forall_neighbours(f,
      if (!f->self)
        S(
        fprintf(stderr,"[%lu]\t%u.%u.%u.%u:%u (%u)\n",
                th_id,
                ((unsigned char *)&f->addr.s_addr)[0],
                ((unsigned char *)&f->addr.s_addr)[1],
                ((unsigned char *)&f->addr.s_addr)[2],
                ((unsigned char *)&f->addr.s_addr)[3],
                f->port,f->status);
        );
    );
    S(fprintf(stderr,"[%lu]\tnew friend list:\n",th_id));
    forall_friends(f,
      S(
      fprintf(stderr,"[%lu]\t%u.%u.%u.%u:%u (%u) %s\n",
              th_id,
              ((unsigned char *)&f->addr.s_addr)[0],
              ((unsigned char *)&f->addr.s_addr)[1],
              ((unsigned char *)&f->addr.s_addr)[2],
              ((unsigned char *)&f->addr.s_addr)[3],
              f->port,f->status,
              f->from_predecessor ? "PRED" : f->from_successor ? "SUCC" : " ");
      );
    );
#endif

    /*  This is here to prevent a silly complaint of gcc  */
    continue;

  }  /*  Infinite loop  */
  
}



/*
 *  Trigger the creation of finger table, by searching finger of magnitude 0.
 *  The search request is sent to the successor;
 *  see findfinger.c (algorithm) and newfinger.c (how to make the req).
 *  Can only be done if not an isolated peer.
 */
void
finger_create
(pthread_t th_id)
{

  peer_t               *f,*ff;
  unsigned int          i,scale_fact;
  neblo_addr            dist;
  neblo_addr            cutoff;
  char                 *buf;
  int                   rc;
  unsigned int          dummy;


  dummy = th_id;

  f = successor();
  if (f == NULL) {

#ifdef DEBUG_R
    S(fprintf(stderr,"[%lu]\tpeer is isolated so cannot have fingers yet\n",
              th_id));
#endif

    goto clean;
  }

  /*
   * Ask to successor to find our finger of magnitude 0, F(0).
   * If successor is beyond CUTOFF distance, adjust the scale factor so
   * as to make F(0) equivalent to F(n) for some n>0 (n is the scale factor).
   * An alternative would be to adjust CUTOFF, but then we would have to
   * manage different values of CUTOFF among different peers...
   */

  dist = _distance_(ring->my_id-ring->smoke,ring->up_bound+1);

  if (dist < _distance_(ring->my_id,ring->up_bound+1)) {
    /*
     *  This peer's addr intvl is so large that with the smoke the lower
     *  bound of the intvl wraps around.  No need for fingers.
     */

#ifdef DEBUG_R
    S(fprintf(stderr,
              "[%lu]\taddr intvl+smoke is too large (%llx) to need fingers\n",
              th_id,dist));
#endif

    goto clean;
  }

  for (scale_fact=0, cutoff=CUTOFF; cutoff; scale_fact++, cutoff<<=1)
    if (dist < cutoff)
      break;

  if (!cutoff || scale_fact >= NUMFINGERS) {

#ifdef DEBUG_R
    S(fprintf(stderr,
              "[%lu]\taddr intvl is too large (%llx) to need fingers yet\n",
              th_id,dist));
#endif

    goto clean;
  }

  ff = &ring->finger_table[scale_fact];
  _do_(_malloc_(buf,KEYBUFLEN+SLACK),!buf,"finger_create():  buf malloc");

  /*  Be sure no other thread is doing the same.  */

  cmd_handle_lock(&ff->handle);

  /*  Create handle for this recursive request.  */
  cmd_handle_set(&ff->handle);
  sprintf(buf,"findfinger %u %u %016llx %lu %u %u %s %s",
          0                          /* finger of magnitude 0 */,
          scale_fact,
          (ULL)(ring->my_id-ring->smoke) /* see findfinger.c for the algo */,
          (unsigned long)my_ipaddr.s_addr,PORT,
          cmd_handle_val(&ff->handle),
          pubkeymod,pubkeyexp);
  rc = tell_peer(f,1 /* from pred */,buf,strlen(buf)+1,0,0 /* dummy th_id */);
  if (rc < 0) {
    /*  Failed to emit "findfinger" req.  Try later.  */
    cmd_handle_reset(&ff->handle);
  } else {

#ifdef DEBUG_M
    S(fprintf(stderr,"[%lu]\t<< %s\n",th_id,buf));
#endif

  }

  cmd_handle_unlock(&ff->handle);
  _free_(buf);
  goto exit;

clean:

  pthread_mutex_lock(&clean_lock);
  if (!dont_clean)
    for (i=0; i<NUMFINGERS; i++)
      if (!pend_read_ref(&ring->finger_table[i]))
        clean_peer_desc(&ring->finger_table[i]);
  pthread_mutex_unlock(&clean_lock);

exit:

  return;

}



/*
 * neblo_check_fingers.
 *
 * Periodically triggers the rebuilding of finger table.
 *
 */

void *
neblo_check_fingers(void *data) 
{

  pthread_t             th_id;
  char                 *cmd,*rsp;
#ifdef DEBUG_R
  unsigned int          i;
  peer_t               *f;
#endif
  struct timeval        t1,t2;
#if 0
  int                   rc;
#endif


  th_id = *(pthread_t *)data;

  _do_(_malloc_(cmd,BUFLEN+SLACK),!cmd,"neblo_check_fingers(): cmd malloc");
  _do_(_malloc_(rsp,BUFLEN+SLACK),!rsp,"neblo_check_fingers(): rsp malloc");

  sprintf(cmd,"ping");
  
  gettimeofday(&t1,NULL);

  for (;;) {
    
    /*
     *  Run periodically, insensitive to the time spent to carry out the
     *  checks.
     */
    gettimeofday(&t2,NULL);
    if ((unsigned int)(t2.tv_sec - t1.tv_sec) < CHK_PERIOD_FINGERS)
      sleep(CHK_PERIOD_FINGERS - (t2.tv_sec - t1.tv_sec));
    gettimeofday(&t1,NULL);

    if (!ring->enabled) continue;

#ifdef DEBUG_R
    S(
    fprintf(stderr,"[%lu]\tneblo_check_fingers()\n",th_id);
    fprintf(stderr,"[%lu]\tcurrent finger table:\n",th_id);
    );
    pthread_mutex_lock(&ring->lock);
    for (i=0; i<NUMFINGERS; i++) {
      f = &ring->finger_table[i];
      if (!(f->status & (OK | TOCALL | PROBABLYDEAD)))
        continue;
      S(
      fprintf(stderr,"\tmagnitude %u (presumable dist. %016llx): %u.%u.%u.%u:%u (%u)\n",
                i,finger_distance_rule(i),
                ((unsigned char *)&f->addr.s_addr)[0],
                ((unsigned char *)&f->addr.s_addr)[1],
                ((unsigned char *)&f->addr.s_addr)[2],
                ((unsigned char *)&f->addr.s_addr)[3],
                f->port,f->status);
      );
    }
    pthread_mutex_unlock(&ring->lock);
#endif

    /*  This is unnecessary: fingers are periodically rebuilt  */
#if 0
    /*  loop over the finger table  */

    for (i=0; i<NUMFINGERS; i++) {

      f = &ring->finger_table[i];

      if (!(f->status & (OK | TOCALL | PROBABLYDEAD))) continue;

#ifdef DEBUG
      S(fprintf(stderr,"[%lu]\tcheck finger %u.%u.%u.%u:%u...\n",
                th_id,
                ((unsigned char *)&f->addr.s_addr)[0],
                ((unsigned char *)&f->addr.s_addr)[1],
                ((unsigned char *)&f->addr.s_addr)[2],
                ((unsigned char *)&f->addr.s_addr)[3],
                f->port));
#endif

      rc = tell_peer(f,0,cmd,strlen(cmd)+1,0,th_id);
      if (rc >= 0) {
        bzero(rsp,4);
        rc = listen_peer(f,rsp,strlen("ping")+1,TIMEOUT,0);
        if (rc >= 0)
          rc = strncmp(rsp,cmd,strlen(cmd));
          rc = -(rc != 0);  /*  -1 iff strings differ from each other  */
      }
      if (rc >= 0) {
        /* finger OK, go next */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\tOK\n",th_id));
#endif

        continue;
      }
      if (rc == -2) {
        /* finger is dead and its status has been set to DEAD. go next */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        continue;
      }
      /* finger has troubles but unsure is dead */
      if (!(f->status & PROBABLYDEAD)) {
        /*  countdown to death begins...  */
        f->status = PROBABLYDEAD;
        f->deadline = PEER_DEADLINE;
        if (f->deadline) {

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\tPROBABLYDEAD, deadline %u\n",
                  th_id,f->deadline));
#endif

        } else {
          f->status = DEAD;

#ifdef DEBUG
          S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        }
      } else {
        /*  countdown to death proceeds...  */
        if (!--f->deadline) {
          /* troubles for too long a time. assume it is dead */
          f->status = DEAD;

#ifdef DEBUG
          S(fprintf(stderr,"[%lu]\tDEAD\n",th_id));
#endif

        }

#ifdef DEBUG
        else
          S(fprintf(stderr,"[%lu]\tPROBABLYDEAD, deadline %u\n",
                            th_id,f->deadline));
#endif

      }

    }    /*  for all fingers  */
#endif

    /*
     * purge the list of old fingers from elements no longer referenced
     * by any pending "read"
     */
    forall_oldfingers(f, if (!pend_read_ref(f)) f->status = DEAD);

    purge_peer_list(&ring->oldfinger_list,&ring->oldfinger_list_tail,
                    &ring->oldfinger_lock);

    /*  Trigger rebuilding of the finger table  */
    finger_create(th_id);

  }  /*  Infinite loop  */

}
